#include "framebuffer.h"
#include "v3.h"
#include "math.h"
#include <iostream>
#include "scene.h"

FrameBuffer::FrameBuffer(int u0, int v0, 
  int _w, int _h) : Fl_Gl_Window(u0, v0, _w, _h, 0) {

  w = _w;
  h = _h;
  pix = new unsigned int[w*h];
  zb = new float[w*h];
  Set(0xFFFFFFFF);
  isHW = false;

}


void FrameBuffer::draw() {

  if (isHW) {
//    scene->RenderGL();
    scene->RenderGPU();
  }
  else {
    glDrawPixels(w, h, GL_RGBA, 
      GL_UNSIGNED_BYTE, pix);
  }

}

void FrameBuffer::Set(unsigned int color) {

  for (int i = 0; i < w*h; i++)
    pix[i] = color;

}

void FrameBuffer::FrameSetup(unsigned int color) {

  Set(color);
  for (int i = 0; i < w*h; i++)
    zb[i] = 0.0f;

}


void FrameBuffer::Set(int u, int v, unsigned int col) {

  pix[(h-1-v)*w+u] = col;

}

void FrameBuffer::SetGuarded(int u, int v, unsigned int col) {

  if (u < 0 || u > w-1 || v < 0 || v > h-1)
    return;
  Set(u, v, col);

}

void FrameBuffer::DrawPoint(int u, int v, int psize, 
                            unsigned int color) {

  for (int vv = v-psize/2; vv <= v+psize/2; vv++) {
    for (int uu = u-psize/2; uu <= u+psize/2; uu++) {
      SetGuarded(uu, vv, color);
    }
  }

}

void FrameBuffer::DrawCircle(float uf, float vf, 
  float r, unsigned int color) {

  AABB bb;
  bb.AddPoint(V3(-r+uf, -r+vf, 0.0f));
  bb.AddPoint(V3(r+uf, r+vf, 0.0f));

  if (!ClipBB(bb))
    return;

  int lefti = (int) (bb[0][0]+0.5f);
  int righti = (int) (bb[1][0]-0.5f);
  int topi = (int) (bb[0][1]+0.5f);
  int bottomi = (int) (bb[1][1]-0.5f);

  V3 center(uf, vf, 0.0f);
  for (int v = topi; v <= bottomi; v++) {
    for (int u = lefti; u <= righti; u++) {
      V3 pixv(.5f+(float)u, .5f+(float)v, 0.0f);
      if ((center-pixv).Length() > r)
        continue;
      Set(u, v, color);
    }
  }

}

void FrameBuffer::DrawSegment(V3 p0, V3 p1, 
  unsigned int color) {

    V3 c0;
    c0.SetFromColor(color);
    DrawSegment(p0, p1, c0, c0);

}

void FrameBuffer::DrawSegment(V3 p0, V3 p1, 
  V3 c0, V3 c1) {

    float du = fabsf(p0[0]-p1[0]);
    float dv = fabsf(p0[1]-p1[1]);
    int stepsN = (du > dv) ? (int) ceilf(du) : (int) ceilf(dv);
    for (int i = 0; i < stepsN; i++) {
      float intfrac = ((float)i/(float)(stepsN-1));
      V3 p = p0 + (p1-p0)*intfrac;
      int u = (int)p[0];
      int v = (int)p[1];
      V3 c = c0 + (c1-c0)*intfrac;
      SetGuarded(u, v, c.GetColor());
    }
}

FrameBuffer* FrameBuffer::FindEdges() {

  FrameBuffer *ret;
  ret = new FrameBuffer(0, 0, w, h);
  ret->label("Edge map");

  ret->Set(0xFF000000);
  //float krn[3][3] = {
  //  {0.0f, -1.0f, 0.0f}, 
  //  {-1.0f, 4.0f, -1.0f}, 
  //  {0.0f, -1.0f, 0.0f}};
  float krn[3][3] = {
    {1.0f, 2.0f, 1.0f}, 
    {2.0f, 4.0f, 2.0f}, 
    {1.0f, 2.0f, 1.0f}};
  float sumk = 16.0f; 
  for (int v = 1; v < h-1; v++) {
    for (int u = 1; u < w-1; u++) {
      V3 currv(0.0f, 0.0f, 0.0f);
      for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
          currv = currv + Getv(u-1+j, v-1+i)*krn[i][j];
        }
      }
      //currv = V3(fabsf(currv[0]), fabsf(currv[1]), 
      //  fabsf(currv[2]));
      //currv = currv * 0.25f;
      currv = currv / sumk;
      ret->Set(u, v, currv.GetColor());
    }
  }

  return ret;

}

unsigned int FrameBuffer::Get(int u, int v) {

  return pix[(h-1-v)*w+u];

}

V3 FrameBuffer::Getv(int u, int v) {

  V3 ret;
  ret.SetFromColor(Get(u, v));
  return ret;

}

bool FrameBuffer::ClipBB(AABB &bb) {

  if (bb[0][0] > (float)w)
    return false;
  if (bb[1][0] < 0.0f)
    return false;
  if (bb[0][1] > (float)h)
    return false;
  if (bb[1][1] < 0.0f)
    return false;

  if (bb[0][0] < 0.0f)
    bb[0][0] = 0.0f;
  if (bb[0][1] < 0.0f)
    bb[0][1] = 0.0f;

  if (bb[1][0] > (float)w)
    bb[1][0] = (float)w;
  if (bb[1][1] > (float)h)
    bb[1][1] = (float)h;

  return true;

}

void FrameBuffer::DrawSegment3D(V3 p0, V3 p1, 
  V3 c0, V3 c1, PPC *ppc) {

  V3 pp0, pp1;
  if (!ppc->Project(p0, pp0) || !ppc->Project(p1, pp1))
      return;

  DrawSegment(pp0, pp1, c0, c1);

}


void FrameBuffer::DrawAABB3D(AABB &aabb, V3 col[2], PPC *ppc) {

  DrawSegment3D(
    V3(aabb.corners[0][0], aabb.corners[0][1],aabb.corners[0][2]),  
    V3(aabb.corners[1][0], aabb.corners[0][1],aabb.corners[0][2]),
    col[1], col[1], ppc);
  DrawSegment3D(
    V3(aabb.corners[1][0], aabb.corners[0][1],aabb.corners[0][2]),
    V3(aabb.corners[1][0], aabb.corners[1][1],aabb.corners[0][2]),
    col[1], col[1], ppc);
  DrawSegment3D(
    V3(aabb.corners[1][0], aabb.corners[1][1],aabb.corners[0][2]),
    V3(aabb.corners[0][0], aabb.corners[1][1],aabb.corners[0][2]),
    col[1], col[1], ppc);
  DrawSegment3D(
    V3(aabb.corners[0][0], aabb.corners[1][1],aabb.corners[0][2]),
    V3(aabb.corners[0][0], aabb.corners[0][1],aabb.corners[0][2]),
    col[1], col[1], ppc);

  DrawSegment3D(
    V3(aabb.corners[0][0], aabb.corners[0][1],aabb.corners[1][2]),  
    V3(aabb.corners[1][0], aabb.corners[0][1],aabb.corners[1][2]),
    col[0], col[0], ppc);
  DrawSegment3D(
    V3(aabb.corners[1][0], aabb.corners[0][1],aabb.corners[1][2]),
    V3(aabb.corners[1][0], aabb.corners[1][1],aabb.corners[1][2]),
    col[0], col[0], ppc);
  DrawSegment3D(
    V3(aabb.corners[1][0], aabb.corners[1][1],aabb.corners[1][2]),
    V3(aabb.corners[0][0], aabb.corners[1][1],aabb.corners[1][2]),
    col[0], col[0], ppc);
  DrawSegment3D(
    V3(aabb.corners[0][0], aabb.corners[1][1],aabb.corners[1][2]),
    V3(aabb.corners[0][0], aabb.corners[0][1],aabb.corners[1][2]),
    col[0], col[0], ppc);

  DrawSegment3D(
    V3(aabb.corners[0][0], aabb.corners[0][1],aabb.corners[0][2]),  
    V3(aabb.corners[0][0], aabb.corners[0][1],aabb.corners[1][2]),  
    col[1], col[0], ppc);
  DrawSegment3D(
    V3(aabb.corners[0][0], aabb.corners[1][1],aabb.corners[0][2]),  
    V3(aabb.corners[0][0], aabb.corners[1][1],aabb.corners[1][2]),  
    col[1], col[0], ppc);
  DrawSegment3D(
    V3(aabb.corners[1][0], aabb.corners[1][1],aabb.corners[0][2]),  
    V3(aabb.corners[1][0], aabb.corners[1][1],aabb.corners[1][2]),  
    col[1], col[0], ppc);
  DrawSegment3D(
    V3(aabb.corners[1][0], aabb.corners[0][1],aabb.corners[0][2]),  
    V3(aabb.corners[1][0], aabb.corners[0][1],aabb.corners[1][2]),  
    col[1], col[0], ppc);

}


void FrameBuffer::DrawPoint3D(V3 P, int psize, 
  unsigned int color, PPC *ppc) {

    V3 pP;
    if (!ppc->Project(P, pP))
      return;
    DrawPoint((int)pP[0], (int)pP[1], psize, color);

}

void FrameBuffer::SetZB(float currz, int u, int v) {

  int uv = (h-1-v)*w+u;
  zb[uv] = currz;

}


bool FrameBuffer::IsCloser(float currz, int u, int v) {

  int uv = (h-1-v)*w+u;
  return (currz > zb[uv]);

}

bool FrameBuffer::IsVisible(float currz, int u, int v) {

  int uv = (h-1-v)*w+u;
  float eps = 0.1f;
  return (currz > zb[uv] - eps);

}

int FrameBuffer::handle(int event)  {  

  switch(event) 
  {   
  case FL_KEYBOARD: {
    KeyboardHandle();
    return 0;
  }
  default:
    break;
  }
  return 0;

}


void FrameBuffer::KeyboardHandle() {

  float tstep = 3.0f;
  float rstep = 3.0f;
  int key = Fl::event_key();
  switch (key) {
    case FL_Left: {
      scene->ppc->Pan(rstep);
      scene->Render();
      break;
    }
    case FL_Right: {
      scene->ppc->Pan(-rstep);
      scene->Render();
      break;
    }
    case 'q': {
      scene->ppc->Tilt(rstep);
      scene->Render();
      break;
    }
    case 'w': {
      scene->ppc->Tilt(-rstep);
      scene->Render();
      break;
    }
    case 'o': {
      scene->ppc->Roll(-rstep);
      scene->Render();
      break;
    }
    case 'p': {
      scene->ppc->Roll(rstep);
      scene->Render();
      break;
    }
    case FL_Up: {
      scene->ppc->TranslateFW(tstep);
      scene->Render();
      break;
    }
    case FL_Down: {
      scene->ppc->TranslateFW(-tstep);
      scene->Render();
      break;
    }
    case ',': {
      V3 trans = scene->ppc->a.Normalized()*tstep;
      scene->ppc->Translate(trans*-1.0f);
      scene->Render();
      break;
    }
    case '.': {
      V3 trans = scene->ppc->a.Normalized()*tstep;
      scene->ppc->Translate(trans);
      scene->Render();
      break;
    }
    case '1': {
      V3 trans = scene->ppc->b.Normalized()*tstep;
      scene->ppc->Translate(trans*-1.0f);
      scene->Render();
      break;
    }
    case '2': {
      V3 trans = scene->ppc->b.Normalized()*tstep;
      scene->ppc->Translate(trans);
      scene->Render();
      break;
    }
    case 'z': {
      scene->ppc->ChangeFocalLength(1.25f);
      scene->Render();
      break;
    }
    case 'x': {
      scene->ppc->ChangeFocalLength(1.0f/1.25f);
      scene->Render();
      break;
    }
    default:
      cerr << "INFO: do not understand keypress" << endl;
  }
}

void FrameBuffer::SetChecker(unsigned int cols[2], int cw) {

  for (int v = 0; v < h; v++) {
    for (int u = 0; u < w; u++) {
      int uc = u / cw;
      int vc = v / cw;
      unsigned int col;
      if ((uc + vc)%2) {
        col = cols[1];
      }
      else {
        col = cols[0];
      }
      Set(u, v, col);
    }
  }

}

unsigned int FrameBuffer::GetST(float s, float t) {

  bool mirror = true;
  int si = (int)s;
  float sf = s - (float)si;
  int ti = (int)t;
  float tf = t - (float)ti;
  int u, v;
  if (!mirror) {
    u = (int) (sf*(float)w);
    v = (int) (tf*(float)h);
  }
  else {
    if (si%2 == 0)
      u = (int) (sf*(float)w);
    else
      u = (int) ((1.0f-sf)*(float)w);
    if (ti%2 == 0)
      v = (int) (tf*(float)h);
    else
      v = (int) ((1.0f-tf)*(float)h);
  }
  return Get(u, v);

}
